# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# IMPORTANT!! ######################################################################################
# please note that this implementation must not be understood as valid for a production environment.
# It is provided for testing purposes only and to provide a basic reference interface
####################################################################################################
import json
import sqlite3

import logging
logger = logging.getLogger(__name__)

from krules_core.subject import PropertyType, SubjectExtProperty, SubjectProperty


class SQLLiteSubjectStorage(object):

    def __init__(self, subject, dbfile):
        self._dbfile = dbfile
        self._subject = subject
        self._conn = sqlite3.connect(dbfile)

        create_subjects_table_sql = """
            CREATE TABLE IF NOT EXISTS subjects (
                subject TEXT NOT NULL,
                property TEXT NOT NULL,
                proptype TEXT NOT NULL,
                propvalue TEXT,
                PRIMARY KEY (subject, property, proptype)
                ) WITHOUT ROWID
        """
        c = self._conn.cursor()
        c.execute(create_subjects_table_sql)
        c = self._conn.cursor()
        c.execute("CREATE INDEX IF NOT EXISTS idx_subjects ON subjects(subject)")
        self._close_connection()

    def __str__(self):

        dbtype = "onfile"
        if self._dbfile == ":memory:":
            dbtype = "inmemory"

        return "sqlite_{}_{}".format(dbtype, self._subject)

    def is_concurrency_safe(self):
        return self._dbfile != ":memory:"

    def is_persistent(self):
        return self._dbfile != ":memory:"

    def _get_connection(self):

        if self._dbfile == ":memory:":
            return self._conn
        self._conn.close()
        self._conn = sqlite3.connect(self._dbfile)
        return self._conn

    def _close_connection(self):
        # does not close connection if database is not persistent
        if self._dbfile != ":memory:":
            self._conn.close()

    def load(self):

        select_props_sql = """
            SELECT property, proptype, propvalue
            FROM subjects
            WHERE subject = ?
        """
        conn = self._get_connection()
        c = conn.cursor()
        c.execute(select_props_sql, (self._subject,))

        res = {
            PropertyType.DEFAULT: {},
            PropertyType.EXTENDED: {}
        }
        rows = c.fetchall()
        for row in rows:
            res[row[1]][row[0]] = json.loads(row[2])

        self._close_connection()

        return res[PropertyType.DEFAULT], res[PropertyType.EXTENDED]

    def store(self, inserts=[], updates=[], deletes=[]):

        conn = self._get_connection()

        sql_script = ""

        # inserts
        for prop in inserts:
            sql_script += "INSERT INTO subjects (subject, property, proptype, propvalue) "\
                          "VALUES ('{}', '{}', '{}', '{}');\n".format(
                self._subject, prop.name, prop.type, prop.json_value().replace("'", "''")
            )

        # updates
        for prop in updates:
            sql_script += "UPDATE subjects SET propvalue='{}' WHERE "\
                          "subject = '{}' "\
                          "AND property = '{}' "\
                          "AND proptype = '{}';".format(
                prop.json_value().replace("'", "''"), self._subject, prop.name, prop.type
            )

        # deletes
        for prop in deletes:
            sql_script += "DELETE FROM subjects WHERE " \
                          "subject = '{}' " \
                          "AND property = '{}' " \
                          "AND proptype = '{}';".format(
                self._subject, prop.name, prop.type
            )

        # TODO: when fails shoud a failback function should be called
        #   checking at least each insert property if needs update instead
        conn.executescript(sql_script)

        self._close_connection()

    def set(self, prop, old_value_default=None):
        """
        Set value for property, works both in update and insert
        Returns old value
        Note that for this implementation a write lock is required on the entire database,
        a row-level lock would be preferable
        """
        conn = self._get_connection()
        #c = conn.cursor()
        conn.execute("BEGIN IMMEDIATE")

        res = conn.execute("SELECT propvalue FROM subjects WHERE subject=? and property=? and proptype=?",
                        (self._subject, prop.name, prop.type)).fetchall()

        old_value = old_value_default

        try:
            if len(res):
                old_value = json.loads(res[0][0])

                conn.execute("UPDATE subjects SET propvalue=? WHERE subject=? and property=? and proptype=?",
                          (prop.json_value(old_value), self._subject, prop.name, prop.type))
            else:
                conn.execute("INSERT INTO subjects (subject, property, proptype, propvalue) "\
                              "VALUES ('{}', '{}', '{}', '{}');\n".format(
                    self._subject, prop.name, prop.type, prop.json_value(old_value_default))
                )
        except Exception as ex:
            conn.execute("ROLLBACK")
            self._close_connection()
            raise ex

        conn.execute("COMMIT")
        self._close_connection()

        new_value = prop.get_value(old_value)
        return new_value, old_value

    def get(self, prop):
        """
        Get a single property
        Raises AttributeError if not found
        """
        conn = self._get_connection()
        res = conn.execute("SELECT propvalue FROM subjects WHERE subject=? and property=? and proptype=?",
                        (self._subject, prop.name, prop.type)).fetchall()
        if not len(res):
            self._close_connection()
            raise AttributeError(prop.name)
        self._close_connection()
        return json.loads(res[0][0])

    # def incr(self, prop, amount=1):
    #     """
    #     some backends may have specialized functions for this operation (eg: redis)
    #     """
    #     prop.value = lambda x: x is None and 0 + amount or x + amount
    #     return self.set(prop)
    #
    # def decr(self, prop, amount=1):
    #     """
    #     some backends may have specialized functions for this operation (eg: redis)
    #     """
    #     prop.value = lambda x: x is None and 0 - amount or x - amount
    #     return self.set(prop)

    def delete(self, prop):
        """
        Delete a single property
        """
        conn = self._get_connection()

        conn.execute("DELETE FROM subjects WHERE subject = ? AND property = ? AND proptype = ?",
                     (self._subject, prop.name, prop.type))

        conn.commit()

        self._close_connection()

    def get_ext_props(self):

        conn = self._get_connection()

        props = conn.execute("SELECT property, propvalue from subjects WHERE proptype = ? and subject = ?",
                             (PropertyType.EXTENDED, self._subject)).fetchall()
        #res = [SubjectExtProperty(pname, json.loads(pvalue)) for pname, pvalue in props]
        self._close_connection()

        return dict((k, json.loads(v)) for k, v in props)


    ## WE DON?T NEED IT
    # def get_all_properties(self):
    #     """
    #     Get all properties
    #     """
    #     conn = self._get_connection()
    #     rows = conn.execute("SELECT property, proptype, propvalue FROM subjects WHERE subject = ?", (self._subject,))\
    #         .fetchall()
    #
    #
    #     res = []
    #     for pname, ptype, pvalue in rows:
    #         klass = ptype == PropertyType.DEFAULT and SubjectProperty or SubjectExtProperty
    #         res.append(klass(pname, json.loads(pvalue)))
    #
    #     self._close_connection()
    #
    #     return res

    def flush(self):

        conn = self._get_connection()

        conn.execute("DELETE FROM subjects WHERE subject = '{}';".format(self._subject))

        conn.commit()

        self._close_connection()




# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from krules_core import RuleConst
from krules_core.core import RuleFactory
import os

def load_rules_from_rulesdata(rulesdata):

    description=""
    for el in rulesdata:
        if type(el) == type(""):
            description=el
        elif type(el) == type({}) and RuleConst.RULENAME in el:
            el[RuleConst.DESCRIPTION] = description
            if el.get(RuleConst.SUBSCRIBE_TO, None) is None:
                el[RuleConst.SUBSCRIBE_TO] = os.environ["TOPIC"]  # TODO: should not rely on this variable. The name of the subscription message should be inferred differently
            RuleFactory.create(**el)
            description = ""



# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO unit tests
import inspect


class ExceptionsDumpers(object):

    def __init__(self):
        self._dumpers = {}

    def set(self, dumper):
        strcls = ""
        ex_cls = dumper.register_for
        if type(ex_cls) == str:
            strcls = ex_cls
        else:
            strcls = ".".join([ex_cls.__module__, ex_cls.__name__])
        self._dumpers[strcls] = dumper

    def dump(self, ex):
        strcls = ".".join([type(ex).__module__, type(ex).__name__])
        mro = [".".join([ex_cls.__module__, ex_cls.__name__])
               for ex_cls in inspect.getmro(ex.__class__)]
        mro.reverse()
        mro.extend([strcls])
        dumped = {}
        for cls in mro:
            if cls in self._dumpers:
                dumped.update(self._dumpers[cls].dump(ex))
        return dumped


class ExceptionDumperBase(object):

    register_for = 'builtins.Exception'

    @staticmethod
    def dump(ex):
        return {"args": ex.args}


## common exceptions dumpers

# python requests HTTPError
class RequestsHTTPErrorDumper(object):

    register_for = 'requests.exceptions.HTTPError'

    @staticmethod
    def dump(ex):
        return {
            "response_code": ex.response.status_code,
            "response_text": ex.response.text
        }







# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from rx import subject as rx_subject

import dependency_injector.providers as providers
from krules_core.base_functions import Callable

from krules_core.route.dispatcher import BaseDispatcher

from krules_core.core import RuleFactory

from krules_core import RuleConst

from krules_core.providers import (
    event_router_factory,
    proc_events_rx_factory,
    subject_factory,
    event_dispatcher_factory
)


def _assert(expr, msg="test failed"):
    assert expr, msg
    return True


counter = 0


@pytest.fixture
def subject():
    global counter
    counter += 1

    return subject_factory('test-subject-{0}'.format(counter)).flush()


@pytest.fixture
def router():
    router = event_router_factory()
    router.unregister_all()
    proc_events_rx_factory.queue.clear()

    return event_router_factory()


def test_internal_routing(subject, router):
    RuleFactory.create('test-rule-filters-pass',
                       subscribe_to="test-type",
                       data={
                           RuleConst.FILTERS: [
                               Callable(
                                   lambda self: True
                               ),
                           ],
                           RuleConst.PROCESSING: [
                               Callable(
                                   lambda self:
                                   self.payload.setdefault(RuleConst.PASSED, True)
                               ),
                           ],
                       })

    RuleFactory.create('test-rule-filters-fails',
                       subscribe_to="test-type",
                       data={
                           RuleConst.FILTERS: [
                               Callable(lambda self: False),
                               Callable(lambda self: True),
                           ],
                           RuleConst.PROCESSING: [
                               Callable(
                                   lambda self:
                                   self.payload.setdefault(RuleConst.PASSED, False)
                               ),
                           ],
                       })

    proc_events_rx_factory.subscribe(
        lambda x: x[RuleConst.RULENAME] == 'test-rule-filters-pass' and
                  _assert(
                      x[RuleConst.PASSED] and
                      len(x[RuleConst.PROCESSING]) > 0 or print("##### LEN ", x[RuleConst.PROCESSING])
                  )
    )
    proc_events_rx_factory.subscribe(
        lambda x: x[RuleConst.RULENAME] == 'test-rule-filters-fails' and
                  _assert(
                      not x[RuleConst.PASSED] and
                      len(x[RuleConst.PROCESSING]) == 0
                  )
    )

    router.route("test-type", subject, {})


def test_dispatch(subject, router):
    _dispatched_events = []

    class _TestDispatcher(BaseDispatcher):

        def dispatch(self, event_type, subject, payload, **extra):
            _dispatched_events.append((event_type, subject, payload))

    event_dispatcher_factory.override(
        providers.Singleton(lambda: _TestDispatcher())
    )

    router.route('test-unhandled-event', subject, {"data": 1})

    t, subject, payload = _dispatched_events.pop()
    _assert(
        t == 'test-unhandled-event' and
        subject.name == subject.name and
        payload.get("data") == 1
    )



# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from rx import subject, from_future

from dependency_injector import providers as providers
from krules_core.subject.empty_storage import EmptySubjectStorage

from .route.dispatcher import BaseDispatcher
from .route.router import EventRouter
from .subject.storaged_subject import Subject
from .exceptions_dumpers import ExceptionsDumpers


configs_factory = providers.Singleton(lambda: {})

# for testing/development only
subject_storage_factory = providers.Factory(lambda *args, **kwargs: EmptySubjectStorage())

subject_factory = providers.Factory(Subject)
# proc_events_rx_factory = providers.Singleton(subject.ReplaySubject)
proc_events_rx_factory = subject.ReplaySubject()
event_router_factory = providers.Singleton(EventRouter)
event_dispatcher_factory = providers.Singleton(BaseDispatcher)
exceptions_dumpers_factory = providers.Singleton(ExceptionsDumpers)



# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from krules_core.base_functions import RuleFunctionBase


class PyCall(RuleFunctionBase):
    """
    *Execute a generic python function.
    Differently from* `Process <https://intro.krules.io/Processing.html#krules_core.base_functions.processing.Process>`_
    *, which does not allow you to interact with function result, with this function is possible to handle execution
    results whether it succeeds, using a functions which takes execution returned value as unique argument, or throws
    an exception, passing it to a chosen callable, developing a different logic for the two different cases.
    By default an exception during execution does not stop ruleset execution, by the way it is possible to
    propagate any raised exception, setting* **raise_on_error** *to True.*

    ::

        def post_to_api(url, data):
            resp = requests.post(url, data)
            resp.raise_for_status()
            return resp

        # ...

        rulesdata = [
            {
                # We register a device with a post request to an API service and we expect to receive all registration
                # info and the generated device uid.
                # If call failed we dispatch the event again, after 3 fails we raise the call exception blocking ruleset execution.
                rulename: "on-device-onboarding-post-to-api",
                subscibre_to: "device-onboarded",
                ruledata: {
                    processing: [
                        PyCall(
                            post_to_api,
                            kwargs={
                                "url": "https://my_awesome_api/devices",
                                "data": lambda payload: payload["device_info"],
                            },
                            on_success=lambda self:
                                lambda ret:
                                    self.router.route(
                                        subject="device-{}".format(ret.pop("uid"))
                                        event_type="device-posted",
                                        payload=ret
                                    ) # on success we dispatch device-posted event passing API returned data in payload
                            on_error=lambda self:
                                lambda exc:
                                    self.router.route(
                                        event_type="device-onboarded"
                                        payload={
                                            "device_info": self.payload["device_info"],
                                            "retry_count": self.payload.get("retry_count", 0) + 1 # on error we update retry_count in payload before to resend the event
                                        }
                                    )
                            payload_dest="api_response",
                            raise_on_error=lambda payload: payload.get("retry_count", 0) > 2, # if retry count is greater than 2 we block ruleset execution
                        )
                    ]
                }
            },

            # ...
            # Here we handle payload["api_response"] content whether it was successful or not

        ]

    """

    def execute(self, func, args=(), kwargs={}, on_success=None, on_error=None,
                payload_dest="pycall_returns", raise_on_error=True):
        """
        Args:
            func: Function which will be to be executed.
            args: *args that will be pass to **func**
            kwargs: **kwargs that will be pass to **func**
            on_success: Callable to handle **func** result which receive returned value as unique parameters.
            on_error: Callable to handle **func** errors, if func raises an exception this function is called receiving the raised exception.
            payload_dest: Payload's key in which func return value will be stored.
            raise_on_error: Boolean value which indicates if func exceptions will be propagated
        """

        try:
            ret = func(*args, **kwargs)
            if on_success is not None:
                on_success(ret)
            self.payload[payload_dest] = ret
        except Exception as exc:
            if on_error is not None:
                on_error(exc)
            if raise_on_error:
                raise exc






import os

from krules_core.tests.subject.sqlite_storage import SQLLiteSubjectStorage

TEST_FNAME="subjectstore_test.sqlite"


from dependency_injector import providers as providers

from krules_core.providers import subject_storage_factory


def test_memorydatabase():

    if os.path.exists(TEST_FNAME):
        os.unlink(TEST_FNAME)

    subject_storage_factory.override(
        providers.Factory(lambda x, **kwargs: SQLLiteSubjectStorage(x, TEST_FNAME))
    )
    assert subject_storage_factory("test-subject").is_persistent()
    assert subject_storage_factory("test-subject").is_concurrency_safe()






# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json

import pytest

import time
from multiprocessing import Process

from krules_core.providers import subject_storage_factory
from krules_core.subject import SubjectProperty, SubjectExtProperty, PropertyType


@pytest.fixture
def storage_subject1():
    return subject_storage_factory("subject1")


@pytest.fixture
def storage_subject2():
    return subject_storage_factory("subject2")


def test_factories(storage_subject1, storage_subject2):

    assert str(storage_subject1) != str(storage_subject2)

    assert str(subject_storage_factory("subject1")) == str(storage_subject1)


def test_property_kinds():

    prop_p = SubjectProperty("p")
    assert prop_p.value is None
    assert prop_p.type == PropertyType.DEFAULT
    prop_e = SubjectExtProperty("p")
    assert prop_e.type == PropertyType.EXTENDED

    # test simple value
    prop = SubjectProperty("p", 1)
    assert prop.json_value() == json.dumps(1)
    assert prop.get_value() == 1
    prop = SubjectExtProperty("p", 2)
    assert prop.get_value() == 2
    assert prop.json_value() == json.dumps(2)

    # callable no args
    prop = SubjectProperty("p", lambda: 10)
    assert prop.json_value() == json.dumps(10)
    assert prop.get_value() == 10
    # callable with args
    prop = SubjectProperty("p", lambda x: x*2)
    assert prop.json_value(2) == json.dumps(4)
    assert prop.get_value() == 4
    prop = SubjectExtProperty("p", lambda x: x.format(2))
    assert prop.get_value("value is {}") == "value is 2"
    assert prop.json_value("value is {}") == json.dumps("value is 2")


def test_load_store_and_flush(storage_subject1):

    props, ext_props = storage_subject1.load()

    assert props == {} and ext_props == {}

    storage_subject1.store(
        inserts=(
            SubjectProperty("p1", 1),
            SubjectProperty("p2", "2'3"),  # sql needs escape
            SubjectExtProperty("px1", 3),
            SubjectExtProperty("p1", "s1")
        )
    )

    props, ext_props = storage_subject1.load()

    assert len(props) == 2
    assert len(ext_props) == 2
    assert props["p1"] == 1
    assert props["p2"] == "2'3"
    assert ext_props["px1"] == 3
    assert ext_props["p1"] == "s1"

    storage_subject1.store(
        updates=(
            SubjectProperty("p2", 3),
            SubjectExtProperty("p1", "s2"),
        ),
        deletes=(
            SubjectProperty("p1"),
            SubjectExtProperty("px1"),
        )
    )

    props, ext_props = storage_subject1.load()

    assert len(props) == 1
    assert len(ext_props) == 1
    assert props["p2"] == 3
    assert ext_props["p1"] == "s2"

    storage_subject1.flush()

    props, ext_props = storage_subject1.load()

    assert props == {} and ext_props == {}


def test_multiple_subjects(storage_subject1, storage_subject2):

    storage_subject1.store(
        inserts=(
            SubjectProperty("p1", 1),
            SubjectProperty("p2", 2),
            SubjectExtProperty("p3", 3),
        )
    )
    storage_subject2.store(
        inserts=(
            SubjectProperty("p1", 1),
            SubjectExtProperty("p2", 2),
        )
    )

    props1, ext_props1 = storage_subject1.load()
    props2, ext_props2 = storage_subject2.load()

    assert len(props1) == 2
    assert len(ext_props1) == 1
    assert len(props2) == 1
    assert len(ext_props2) == 1

    storage_subject1.store(
        updates=(
            SubjectProperty("p1", 10),
            SubjectExtProperty("p3", 30)
        ),
        deletes=(
            SubjectProperty("p2"),
        )
    )

    props1, ext_props1 = storage_subject1.load()
    props2, ext_props2 = storage_subject2.load()

    assert len(props1) == 1
    assert len(props2) == 1
    assert len(ext_props2) == 1

    assert props1["p1"] == 10
    assert props2["p1"] == 1


def test_set_and_get(storage_subject1):

    with pytest.raises(AttributeError):
        storage_subject1.get(SubjectProperty("pset"))

    # simple value
    new_value, old_value = storage_subject1.set(SubjectProperty("pset", 1))
    assert old_value is None
    assert new_value == 1
    storage_subject1.delete(SubjectProperty("pset"))
    new_value, old_value = storage_subject1.set(SubjectProperty("pset", 1), 0)
    assert old_value == 0
    assert new_value == 1
    new_value, old_value = storage_subject1.set(SubjectProperty("pset", "1'2"))
    assert new_value == "1'2"
    assert old_value == 1

    assert storage_subject1.get(SubjectProperty("pset")) == "1'2"

    # computed value
    storage_subject1.set(SubjectProperty("pset", lambda: "1'2"))  # no args
    storage_subject1.set(SubjectProperty("pset", lambda x: x.replace("'", "$")))

    assert storage_subject1.get(SubjectProperty("pset")) == "1$2"

    # having exceptions in writing operations does not cause bad status (eg; write-locked database)
    with pytest.raises(Exception):
        storage_subject1.set(SubjectProperty("pset", lambda x: x/0))
    storage_subject1.set(SubjectProperty("psetX", 0))

    # test concurrency
    if storage_subject1.is_concurrency_safe():
        def _wait_and_incr(val):
            time.sleep(1)
            return val+1

        def _call_wait_and_incr():
            storage_subject1.set(SubjectProperty("psetX", _wait_and_incr))
        p = Process(target=_call_wait_and_incr)
        p.start()
        storage_subject1.set(SubjectProperty("psetX", lambda x: x+1))
        p.join()
        assert storage_subject1.get(SubjectProperty("psetX")) == 2


def test_ext_props(storage_subject1, storage_subject2):
    storage_subject1.flush()
    storage_subject2.flush()

    storage_subject1.set(SubjectProperty("p1", 1))
    storage_subject1.set(SubjectProperty("p2", 2))
    storage_subject1.set(SubjectExtProperty("p3", 3))
    storage_subject1.set(SubjectExtProperty("p4", 4))
    storage_subject2.set(SubjectExtProperty("p5", 5))

    props = storage_subject1.get_ext_props()
    assert len(props) == 2

    assert "p3" in props and props["p3"] == 3
    assert "p4" in props and props["p4"] == 4

# def test_get_all_properties(storage_subject1):
#
#     storage_subject1.flush()
#
#     storage_subject1.set(SubjectProperty("p1", 1))
#     storage_subject1.set(SubjectProperty("p2", 2))
#     storage_subject1.set(SubjectExtProperty("p3", 3))
#     storage_subject1.set(SubjectExtProperty("p4", 4))
#
#     all_props = list(storage_subject1.get_all_properties())
#     assert len(all_props) == 4
#     all_props.sort(key=lambda x: x.name)
#     assert all_props[0].type == PropertyType.DEFAULT and all_props[0].name == "p1" and all_props[0].value == 1
#     assert all_props[1].type == PropertyType.DEFAULT and all_props[1].name == "p2" and all_props[1].value == 2
#     assert all_props[2].type == PropertyType.EXTENDED and all_props[2].name == "p3" and all_props[2].value == 3
#     assert all_props[3].type == PropertyType.EXTENDED and all_props[3].name == "p4" and all_props[3].value == 4








# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

from krules_core.subject import PayloadConst

from krules_core.base_functions import RuleFunctionBase
import inspect


class Filter(RuleFunctionBase):
    """
    *Evaluates a boolean expression and returns its value.*
    *The best way to exploit it is to use it in combination with* `Argument Processors <https://intro.krules.io/ArgumentProcessors.html>`_.

    ::

        rulesdata = [
            {
                rulename: "filtered-rules",
                subscibre_to: "k8s.resource.add",
                ruledata: {
                    filters: [
                        Filter(
                            lambda payload:(
                                "my-label" in payload["object"]["metadata"].get("labels", {})
                            )
                        ),
                    ]
                    processing: [
                        ...
                    ]
                }
            }
        ]
    """

    def execute(self, value):
        """
        Args:
            value: Boolean expression which will be evaluated
        """
        return value


class SubjectNameMatch(RuleFunctionBase):
    """
    *Return True if the subject's name matches the given regular expression*

    ::

        rulesdata = [
            {
                rulename: "on-user-login-do-something",
                subscibre_to: "user-login",
                ruledata: {
                    filters: [
                        SubjectNameMatch(r"^user\|(?P<user_id>.+)", payload_dest="user_id"),
                    ]
                    processing: [
                        ...
                    ]
                }
            }
        ]
    """

    def execute(self, regex, payload_dest="subject_match"):
        """
        Args:
            regex: Regular expression which will be evaluated
            payload_dest: Name of the key in the payload where the value of any groups contained in the expression will be saved saved
        """
        import re
        match = re.search(regex, self.subject.name)
        if match is None:
            return False
        self.payload[payload_dest] = match.groupdict()
        return True


class SubjectNameDoesNotMatch(SubjectNameMatch):
    """
    *Return True if the subject's name does not match the given regular expression*
    """

    def execute(self, regex, **kwargs):  # TODO: best inheritage support (**kwargs satisfy base class signature)
        """
        Args:
            regex: Regular expression which will be evaluated
        """

        return not super().execute(regex)


class CheckSubjectProperty(RuleFunctionBase):
    """
    *Returns True if the given subject property exists and, if provided, match the given value.*

    ::

        rulesdata = [
            {
                rulename: "on-device-status-new-update-do-something",
                subscibre_to: "device-status-updated",
                ruledata: {
                    filters: [
                        CheckSubjectProperty(
                            "last_update",
                            property_value=lambda payload:(
                                lambda value: value > payload.get("update_time")
                            )
                        ),
                    ]
                    processing: [
                        ...
                    ]
                }
            }
        ]
    """

    def execute(self, property_name, property_value=lambda _none_: None, extended=False, use_cache=True):
        """
        Args:
            property_name: The name of the property
            property_value(optional): Could be the expected property value or a callable which takes the actual property value as unique parameter. It is possible to exploit the `Argument Processors <https://intro.krules.io/ArgumentProcessors.html>`_ to get the expected property value using nested lambda (as in exmaple).
            extended: If True, check extended property [default False]
            use_cache: If False it checks the actual value on the storage backend bypassing the cached value [default True]
        """

        if (property_name not in self.subject and not extended) or (
                extended and property_name not in self.subject.get_ext_props()):
            return False

        _get = extended and self.subject.get_ext or self.subject.get
        if inspect.isfunction(property_value):
            sign = inspect.signature(property_value)
            if str(sign) == '(_none_)':
                return True
            return property_value(_get(property_name, use_cache=use_cache))
        return _get(property_name, use_cache=use_cache) == property_value


class PayloadMatch(RuleFunctionBase):
    """
    *Processes the payload with a given jsonpath expression to check its content.*

    ::

        # event payload = {
        #     "user": "admin",
        #     "skills": [{"id": 1, "rating": 85}, {"id": 2, "value": 53}, {"id": 3, "value": 98}]}
        # }

        rulesdata = [
            {
                rulename: "on-admin-skills-updated-with-wrong-filter",
                subscibre_to: "user-skills-updated",
                ruledata: {
                    filters: [
                        PayloadMatch("$.user", "admin"), # return False because with single_match = False match result is always a list
                    ]
                    processing: [
                        ...
                    ]
                }
            },
            {
                rulename: "on-admin-skills-updated-with-correct-filter",
                subscibre_to: "user-skills-updated",
                ruledata: {
                    filters: [
                        PayloadMatch("$.user", "admin", single_match=True), # return True
                    ]
                    processing: [
                        ...
                    ]
                }
            },
            {
                rulename: "on-user-skills-updated-update-qualified-skills",
                subscibre_to: "user-skills-updated",
                ruledata: {
                    filters: [
                        PayloadMatch("$.data[?@.rating>80]", lambda x: len(x) == 2, payload_dest="qualified_skills")
                    ]
                    processing: [
                        ...
                    ]
                }
            },
        ]
    """

    def execute(self, jp_expr, match_value=lambda _none_: None, payload_dest=None, single_match=False):
        """
        Args:
            jp_expr: Jsonpath expression which will be used to process the payload.
            match_value: It can be both the expected value of the jsonpath expression processing or a boolean function that handles the expression result.
            payload_dest: Payload key in which match result will be stored. If None, no result will be saved. [default None]
            single_match: If True produces a single value as result, a list of values otherwise [default False]
        """

        import jsonpath_rw_ext as jp

        matched = False
        fn = jp.match
        if single_match:
            fn = jp.match1

        match = fn(jp_expr, self.payload)
        if match is not None and len(match):
            matched = True

        if payload_dest:
            self.payload[payload_dest] = match

        if inspect.isfunction(match_value):
            sign = inspect.signature(match_value)
            if str(sign) != '(_none_)':
                matched = match_value(match)
        else:
            matched = match == match_value

        return matched


class PayloadMatchOne(PayloadMatch):
    """
    *Extends* `PayloadMatch <https://intro.krules.io/Filters.html#krules_core.base_functions.filters.PayloadMatch>`_

    ::

        # event payload = {
        #     "user": "admin",
        #     "skills": [{"id": 1, "rating": 85}, {"id": 2, "value": 53}, {"id": 3, "value": 98}]}
        # }

        rulesdata = [
            {
                rulename: "on-admin-skills-updated-do-something",
                subscibre_to: "skills-updated",
                ruledata: {
                    filters: [
                        PayloadMatchOne("$.user", "admin"), # return True, it is the same as PayloadMatch("$.user", "admin", single_match=True)
                    ]
                    processing: [
                        ...
                    ]
                }
            }
        ]
    """

    def execute(self, jp_expr, match_value=lambda _none_: None, payload_dest=None, **kwargs):
        """
        Args:
            jp_expr: Jsonpath expression
            payload_dest: Destination key in payload
        """
        return super().execute(jp_expr, match_value, payload_dest, True)


class OnSubjectPropertyChanged(RuleFunctionBase):
    """
    *Specific function to filter* **subject-property-changed** *event.
    This event is produced whenever a subject property changes and its data contains the property name, the new property value and the old one.*

    ::

        ☁️  cloudevents.Event
        Validation: valid
        Context Attributes,
          specversion: 1.0
          type: subject-property-changed
          source: my-ruleset
          subject: foo
          id: bd198e83-9d7e-4e93-a9ae-aa21a40383c6
          time: 2020-06-16T08:16:57.340692Z
          datacontenttype: application/json
        Extensions,
          knativearrivaltime: 2020-06-16T08:16:57.346535873Z
          originid: bd198e83-9d7e-4e93-a9ae-aa21a40383c6
          propertyname: count
          traceparent: 00-d571530282362927f824bae826e1fa36-a52fceb915060653-00
        Data,
          {
            "property_name": "count",
            "old_value": 0,
            "value": 1
          }

    *It is important to note that the property name is also present also in the event's extensions and can therefore be used as a filter in a Trigger to intercept changes only of a specific property.*

    ::

        apiVersion: eventing.knative.dev/v1
        kind: Trigger
        metadata:
          name: on-count-change-trigger
        spec:
          filter:
            attributes:
              propertyname: count
              type: subject-property-changed
          subscriber:
            ref:
              apiVersion: v1
              kind: Service
              name: on-count-change-handler

    *This function allows you to exec your rule only when a specific property changes and, optionally, only for a specific current or previous value of it.*

    ::

        rulesdata = [
            {
                rulename: "on-tick-update-count",
                subscibre_to: "tick",
                ruledata: {
                    filters: [
                        SetSubjectProperty("count", lambda x: x + 1), # Since in this case the event will be handled within the ruleset itself,
                                                                      # it is not necessary to define any trigger to intercept it
                    ]
                    processing: [
                        ...
                    ]
                }
            },
            {
                rulename: "on-new-count-do-something",
                subscibre_to: "subject-property-changed",
                ruledata: {
                    filters: [
                        SubjectPropertyChanged("count"), # Exec processing section each time property count changes
                    ]
                    processing: [
                        ...
                    ]
                }
            },
            {
                rulename: "on-even-count-do-something",
                subscibre_to: "subject-property-changed",
                ruledata: {
                    filters: [
                        SubjectPropertyChanged(
                            "count",
                            value=lambda v: v%2 == 0
                        ), # Exec processing section each time a new even value is assigned to property count
                    ]
                    processing: [
                        ...
                    ]
                }
            },
            {
                rulename: "on-count-passed-from-zero-to-new-value-do-something",
                subscibre_to: "subject-property-changed",
                ruledata: {
                    filters: [
                        SubjectPropertyChanged("count", old_value=0), # Exec processing section each time property count changes and
                                                                      # its old value is equal to 0
                    ]
                    processing: [
                        ...
                    ]
                }
            },
            {
                rulename: "on-count-passed-from-even-value-to-new-value-greater-then-ten",
                subscibre_to: "subject-property-changed",
                ruledata: {
                    filters: [
                        SubjectPropertyChanged(
                            "count",
                            value=lambda v, o: v > 10 && o%2 == 1
                        ), # Exec processing section each time a new value greater than 10 is assigned to property count and its last value was odd
                    ]
                    processing: [
                        ...
                    ]
                }
            },
        ]
    """

    def execute(self, property_name, value=lambda _none_: None, old_value=lambda _none_: None):
        """
        Args:
            property_name: Name of the changed property. Could be a string, a callable with no args which returns the property name or a boolean callable which accepts the property name as unique parameter
            value: Could be the expected value or  a boolean callable which takes as arguments just the new property value or the new value and the previous one. It is possible to exploit the `Argument Processors <https://intro.krules.io/ArgumentProcessors.html>`_ to get the expected property value.
            old_value: Could be the expected previous property value or a boolean callable which takes as arguments the previous property value. It is possible to exploit the `Argument Processors <https://intro.krules.io/ArgumentProcessors.html>`_ to get the expected old property value.
        """

        # property_name
        if inspect.isfunction(property_name):
            sign = inspect.signature(property_name)
            n_args = len(sign.parameters)
            if n_args == 0:
                matched = self.payload[PayloadConst.PROPERTY_NAME] == property_name()
            elif n_args == 1:
                matched = property_name(self.payload[PayloadConst.PROPERTY_NAME])
            else:
                raise TypeError("takes at most two arguments (received {})".format(n_args))
        else:
            matched = self.payload[PayloadConst.PROPERTY_NAME] == property_name

        if not matched:
            return False

        # value
        if inspect.isfunction(value):
            sign = inspect.signature(value)
            if str(sign) != '(_none_)':
                n_args = len(sign.parameters)
                if n_args == 1:
                    matched = value(self.payload[PayloadConst.VALUE])
                elif n_args == 2:
                    args = [self.payload[PayloadConst.VALUE], self.payload[PayloadConst.OLD_VALUE]]  # for IDE happiness
                    matched = value(*args)
                else:
                    raise TypeError("takes at most three arguments (received {})".format(n_args))
        else:
            matched = self.payload[PayloadConst.VALUE] == value

        if not matched:
            return False

        # old_value
        if inspect.isfunction(old_value):
            sign = inspect.signature(old_value)
            if str(sign) != '(_none_)':
                n_args = len(sign.parameters)
                if n_args == 1:
                    matched = old_value(self.payload[PayloadConst.OLD_VALUE])
                else:
                    raise TypeError("takes at most two arguments (received {})".format(n_args))
        else:
            matched = self.payload[PayloadConst.OLD_VALUE] == old_value

        if not matched:
            return False

        return True



# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
from collections.abc import Mapping

from krules_core.route.router import DispatchPolicyConst

from krules_core.base_functions import RuleFunctionBase, Filter


class Process(Filter):
    """

    *Like* `Filter <https://intro.krules.io/Filters.html#krules_core.base_functions.filters.Filter>`_ *evaluate a given expression but does not return it.*
    *The best way to exploit it is to use it in combination with* `Argument Processors <https://intro.krules.io/ArgumentProcessors.html>`_.

    ::

        from krules_env import RULE_PROC_EVENT

        #...

        rulesdata = [
            {
                # Store processed events with Django ORM
                rulename: "processed-rules",
                subscibre_to: RULE_PROC_EVENT,
                ruledata: {
                    processing: [
                        Process(
                            lambda payload:(
                                ProcessedEvent.objects.create(
                                    rule_name=payload["name"],
                                    type=payload["type"],
                                    subject=payload["subject"],
                                    event_info=payload["event_info"],
                                    payload=payload["payload"],
                                    time=payload["event_info"].get("time", datetime.now().isoformat()),
                                    filters=payload["filters"],
                                    processing=payload["processing"],
                                    got_errors=payload["got_errors"],
                                    processed=payload["processed"],
                                    origin_id=payload["event_info"].get("originid", "-")
                                )
                            )
                        ),
                    ]
                }
            }
        ]

    """
    def execute(self, value):
        """

        """
        super().execute(value)


## PAYLOAD FUNCTIONS ##############################################################


class SetPayloadProperties(RuleFunctionBase):
    """
    *Set the given properties in the payload, if some of that already exist will be overridden*

    ::

        rulesdata = [
            {
                rulename: "on-admin-login-update-payload",
                subscibre_to: "user-login",
                ruledata: {
                    filters: [
                        ...
                        # Check if user is admin
                    ]
                    processing: [
                        SetPayloadProperties(  # Definition with a dictionary
                            lambda: **{
                                "has_admin_access": True,
                                "last_login": datetime.now()
                            }
                        )
                    ]
                }
            },
            {
                rulename: "on-user-login-update-payload",
                subscibre_to: "user-login",
                ruledata: {
                    filters: [
                        ...
                        # Check if user has not admin privileges
                    ]
                    processing: [
                        SetPayloadProperties(  # Definition with named arguments
                            has_admin_access=False,
                            last_login=lambda:datetime.now()
                        )
                    ]
                }
            },
            # Thanks to ArgumentProcessor we can use a lambda, without that last_login
            # would be always equal to the Rule instantiation's datetime while we need the execution's one.
        ]
    """

    def execute(self, **kwargs):
        """
        Args:
              **kwargs: Each named paramenter is the key and the value to update with.
        """
        for k, v in kwargs.items():
            self.payload[k] = v


class SetPayloadProperty(SetPayloadProperties):
    """
    *Extends* `SetPayloadProperties <https://intro.krules.io/Processing.html#krules_core.base_functions.processing.SetPayloadProperties>`_
    *expecting a single property to set*

    ::

        rulesdata = [
            {
                rulename: "on-heather-onboarded-set-class",
                subscibre_to: "device-onboarded",
                ruledata: {
                    filters: [
                        ...
                        # Check if device has characteristics of an heather
                    ]
                    processing: [
                        SetPayloadProperty(
                            property_name="device_class",
                            value="heather"
                        )
                    ]
                }
            },
        ]
    """

    def execute(self, property_name, value):
        """
        Args:
            property_name: Name of property which will be to set,
            value: Value to set.
        """
        super().execute(**{property_name: value})


## SUBJECT FUNCTIONS ################################################################


class SetSubjectProperty(RuleFunctionBase):
    """
    *Set a single property of the subject, supporting atomic operation.*
    *By default, the property is reactive unless is muted (muted=True) or extended (extended=True)*

    ::

        rulesdata = [
            {
                rulename: "set-device-class",
                subscibre_to: "device-onboarded",
                ruledata: {
                    filters: [
                        ...
                        # Check if device has characteristics of an heather
                    ]
                    processing: [
                        SetSubjectProperty(
                            property_name="device_class",
                            value="heather"
                        )
                    ]
                }
            },
            {
                rulename: "on-new-checkup-increment-counter",
                subscibre_to: "checkup",
                ruledata: {
                    processing: [
                        SetSubjectProperty(
                            property_name="checkup_cnt",
                            value=lambda x: x is None and 1 or x + 1 # Operation is atomic
                        )
                    ]
                }
            }
        ]
    """
    def execute(self, property_name, value, extended=False, muted=False, use_cache=True):
        """
        Args:
            property_name: Name of the property to set. It may or may not exist
            value: Value to set. It can be a callable and receives (optionally) the current property value.
                If the property does not exist yet, it receives None. Note that value setting is an atomic operation.
            extended: If True set an extended property instead a standard one. [default False]
            muted: If True no subject-property-changed will be raised after property setting. Note that extended
                properties are always muted so, if extended is True, this parameter will be ignored. [default False]
            use_cache: If False store the property value immediately on the storage, otherwise wait for the end of rule execution. [default False]
        """
        if extended:
            fn = lambda v: self.subject.set_ext(property_name, v, use_cache)
        else:
            fn = lambda v: self.subject.set(property_name, v, muted, use_cache)

        return fn(value)


class SetSubjectPropertyImmediately(SetSubjectProperty):
    """
    *Extends* `SetSubjectProperty <https://intro.krules.io/Processing.html#krules_core.base_functions.processing.SetSubjectProperty>`_
    *setting a property directly to the storage without using the cache (* **use_cache=False** *).
    This could be very helpful to avoid concurrency issues by avoiding running into inconsistencies during the execution.
    The extension's aim is to made code more readable.*
    """
    def execute(self, property_name, value, extended=False, muted=False, **kwargs):
        """

        """
        return super().execute(property_name, value, extended=extended, muted=muted, use_cache=False)


class SetSubjectExtendedProperty(SetSubjectProperty):
    """
    *Extends* `SetSubjectProperty <https://intro.krules.io/Processing.html#krules_core.base_functions.processing.SetSubjectProperty>`_
    *setting an extended property of the subject(* **extended=True** *). Note that* **muted** *is not present anymore
    in the arguments because an extended property is always muted.
    The extension's aim is to made code more readable.*
    """
    def execute(self, property_name, value, use_cache=True, **kwargs):
        """

        """
        return super().execute(property_name, value, extended=True, muted=True, use_cache=use_cache)


class SetSubjectProperties(RuleFunctionBase):
    """
    *Set multiple properties in subject from dictionary. This is allowed only by using cache and not for
    extended properties. Each property set in that way is muted but it is possible to unmute some of that using*
    **unmuted** *parameter*

    ::

        rulesdata = [
            {
                rulename: "on-device-oboarded-update",
                subscibre_to: "device-onboarded",
                ruledata: {
                    filters: [
                        ...
                        # Check if device has characteristics of an heather
                    ]
                    processing: [
                        SetSubjectProperties(
                            props=lambda: {
                                "device_class": "heather",
                                "on_boarding_tm": datetime.now(),
                            },
                            unmuted=["heather"]
                        )
                        # Thanks to ArgumentProcessor we can use a lambda, without that on_boarding_tm
                        # would be always equal to the Rule instantiation's datetime while we need the execution's one.
                    ]
                }
            }
        ]
    """

    def execute(self, props, unmuted=[]):
        """
        Args:
            props: The properties to set
            unmuted: List of property names for which emit property changed events
        """
        for name, value in props.items():
            self.subject.set(name, value, muted=name not in unmuted)


class StoreSubject(RuleFunctionBase):
    """
    *Store alla subject properties on the subject storage and then flush the cache.
    Usually this happens at the end of the ruleset execution.*
    """

    def execute(self):
        self.subject.store()


class FlushSubject(RuleFunctionBase):
    """
    *Remove all subject's properties. It is important tho recall that a subject exists while it has at least a property,
    so* **remove all its properties means remove the subject itself**.

    ::

        rulesdata = [
            {
                rulename: "on-user-unsubscribe-delete-subject",
                subscibre_to: "user-unsubscribed",
                ruledata: {
                    processing: [
                        DeleteProfileFromDB(user_id=lambda subject: subject.user_id),
                        FlushSubject()
                    ]
                }
            },
            {
                rulename: "on-onboard-device-store-properties",
                subscribe_to: "onboard-device",
                ruledata: {
                    processing: [
                        FlushSubject(),
                        SetSubjectProperties(lambda payload: payload["data"]),
                        SetSubjectProperty('status', 'READY'),
                    ],
                },
            },
        ]
    """

    def execute(self):
        self.subject.flush()


#####################################################################################


class Route(RuleFunctionBase):
    """
    *Produce an event inside and/or outside the ruleset, for "sending outside" the event we mean to deliver it to the
    dispatcher component.
    By default an event is dispatched outside only if there is no handler defined in the current ruleset.
    However it is possible to change this behavior using* **dispatch_policy**.
    *Available choices are defined in* **krules_core.route.router.DispatchPolicyConst** *as:*
        - **DEFAULT**: *Dispatched outside only when no handler is found in current ruleset;*
        - **ALWAYS**: *Always dispatched outside even if an handler is found and processed in the current ruleset;*
        - **NEVER**: *Never dispatched outside;*
        - **DIRECT**: *Skip to search for a local handler and send outside directly.*

    ::

        from krules_core.route.router.DispatchPolicyConst import DEFAULT, ALWAYS, NEVER, DIRECT
        from krules_core.event_types import SUBJECT_PROPERTY_CHANGED

        # ...

        rulesdata = [
            {
                rulename: "on-device-onboarded-dispatch-added-event",
                subscibre_to: "device-onboarded",
                ruledata: {
                    processing: [
                        # ...
                        # do something with device
                        Route(
                            subject=lambda payload: payload["device_id"],
                            payload=lambda payload: payload["device_data"],
                            event_type="device-added",
                            # no dispatch_policy is provided so will be used the DEFAULT one
                        ),
                    ]
                }
            },
            {
                rulename: "on-position-change-propagate-event",
                subscibre_to: SUBJECT_PROPERTY_CHANGED,
                ruledata: {
                    filters: [
                        OnSubjectPropertyChanged("position")
                    ]
                    processing: [
                        Route(
                            dispatch_policy=DIRECT
                            # In this case we don't specify neither type, nor subject, nor payload.
                            # We use dispatch_policy DIRECT to propagate the received event outside, this increase
                            # efficiency because we want avoid useless check in other rules subscribed to SUBJECT_PROPERTY_CHANGED.
                            #  Note that the rules are processed following the order in which they were defined.
                        )
                    ]
                }
            },
            {
                rulename: "on-temp-change-propagate-event",
                subscibre_to: SUBJECT_PROPERTY_CHANGED,
                ruledata: {
                    filters: [
                        OnSubjectPropertyChanged("temp", value=lambda v: v > 30)
                    ]
                    processing: [
                        Route(
                            event_type="device-overheated"
                            dispatch_policy=ALWAYS
                            # We want to handle device-overheated event both in the current container and outside, for example to send an external notification
                        )
                    ]
                }
            },
            {
                rulename: "on-device-overheated-schedule-check",
                subscribe_to: "device-overheated",
                ruledata: {
                    # ...
                }
            },
        ]
    """

    def execute(self, event_type=None, subject=None, payload=None, dispatch_policy=DispatchPolicyConst.DEFAULT):
        """
        Args:
            event_type: The event type. If None use current processing event type [default None]
            subject: The event subject. If None use the current subject [default None]
            payload: The event payload. If None use the current payload [default None]
            dispatch_policy: Define the event dispatch policy as explained before. [default DispatchPolicyConst.DEFAULT]
        """

        from krules_core.providers import event_router_factory
        if event_type is None:
            event_type = self.event_type
        if subject is None:
            subject = self.subject
        if payload is None:
            payload = self.payload

        self.router.route(event_type, subject, payload, dispatch_policy=dispatch_policy)


class RaiseException(RuleFunctionBase):
    """
    *Force the given exception raising*

    ::

        from .my_custom_exceptions import UnexpectedPayload # supposing we defined a module with custom exceptions

        rulesdata = [
            {
                rulename: "on-unexpected-payload-raise-exception",
                subscibre_to: "device-onboarded",
                ruledata: {
                    filters: [
                        Return(lambda payload: "device_id" not in payload)
                    ]
                    processing: [
                        RaiseException(
                            UnexpectedPayload("device_id missing!")
                        )
                    ]
                }
            },
        ]
    """

    def execute(self, ex):
        """
        Args:
            ex: The exception to be raised
        """

        raise ex




# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import socket

logger = logging.getLogger("__router__")


class DispatchPolicyConst:
    DEFAULT = "default"
    ALWAYS = "always"
    NEVER = "never"
    DIRECT = "direct"


class EventRouter(object):

    def __init__(self):
        self._callables = {}

    def register(self, rule, event_type):
        logger.debug("register {0} for {1}".format(rule, event_type))
        if event_type not in self._callables:
            self._callables[event_type] = []
        self._callables[event_type].append(rule._process)

    def unregister(self, event_type):
        logger.debug("unregister event {}".format(event_type))
        count = 0
        if event_type in self._callables:
            for r in self._callables[event_type]:
                count += 1
            del self._callables[event_type]
        return count

    def unregister_all(self):
        count = 0
        types = tuple(self._callables.keys())
        for event_type in types:
            count += self.unregister(event_type)
        return count

    def route(self, event_type, subject, payload, dispatch_policy=DispatchPolicyConst.DEFAULT):

        if isinstance(subject, str):
            # NOTE: this should have already happened if we want to take care or event info
            from krules_core.providers import subject_factory
            subject = subject_factory(subject, event_data=payload)

        from ..providers import event_dispatcher_factory

        _callables = self._callables.get(event_type, None)

        #        try:
        if not dispatch_policy == DispatchPolicyConst.DIRECT:
            if _callables is not None:
                for _callable in _callables:
                    _callable(event_type, subject, payload)
        #        finally:
        #            subject.store()

        # TODO: unit test (policies)
        if dispatch_policy != DispatchPolicyConst.NEVER and _callables is None \
                and dispatch_policy == DispatchPolicyConst.DEFAULT \
                or dispatch_policy == DispatchPolicyConst.ALWAYS \
                or dispatch_policy == DispatchPolicyConst.DIRECT:
            logger.debug("dispatch {} to {} with payload {}".format(event_type, subject, payload))
            return event_dispatcher_factory().dispatch(event_type, subject, payload)



# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.



from abc import ABCMeta, abstractmethod
#import jsonpath_rw_ext as jp
from krules_core.arg_processors import processors, DefaultArgProcessor
from krules_core.providers import event_router_factory, configs_factory


# class with_payload(object):
#
#     def __init__(self, func):
#         self.func = func
#
#     def __call__(self, payload):
#         self._result = self.func(payload)
#         return self._result
#
#     def __repr__(self):
#         if not hasattr(self, "_result"):
#             return "[not called]"
#         return str(self._result)
#
#     def result(self):
#         return getattr(self, "_result", None)
#
# # TODO: with_payload_jp
#
# # TODO: need testing
# class with_subject(object):
#
#     def __init__(self, func):
#         self.func = func
#
#     def __call__(self, subject):
#         self._result = self.func(subject)
#         return self._result
#
#     def __repr__(self):
#         if not hasattr(self, "_result"):
#             return "[not called]"
#         return str(self._result)
#
#     def result(self):
#         return getattr(self, "_result", None)
#
# # TODO: with_subject_jp
#
# # TODO: test
# class with_self(object):
#
#     def __init__(self, func):
#         self.func = func
#
#     def __call__(self, _class):
#         self._result = self.func(_class)
#         return self._result
#
#     def __repr__(self):
#         if not hasattr(self, "_result"):
#             return "[not called]"
#         return str(self._result)
#
#     def result(self):
#         return getattr(self, "_result", None)


class RuleFunctionBase:

    __metaclass__ = ABCMeta

    subject = object()  # just for the ide happiness

    payload = {}
    event_type = ""
    router = object()
    configs = {}
    rule_name = ""

    def __init__(self, *args, **kwargs):
        self._args = []
        for a in args:
            self._args.append(self._get_arg_processor(a))

        self._kwargs = {}
        for k, v in kwargs.items():
            self._kwargs[k] = self._get_arg_processor(v)


    @staticmethod
    def _get_arg_processor(arg):
        for processor in processors:
            if processor.interested_in(arg):
                return processor(arg)
        return DefaultArgProcessor(arg)

    def _get_args(self, instance):
        args = []
        for processor in self._args:
            args.append(processor.process(instance))
        return tuple(args)

    def _get_kwargs(self, instance):
        kwargs = {}
        for key, processor in self._kwargs.items():
            kwargs[key] = processor.process(instance)
        return kwargs

    @abstractmethod
    def execute(self, *args, **kwargs):
        raise NotImplementedError("execute")


from krules_core.base_functions.filters import *
from krules_core.base_functions.processing import *
from krules_core.base_functions.misc import *


def Callable(_callable):
    return type("_CallableRuleFunction", (RuleFunctionBase,), {"execute": _callable})



# Copyright 2019 The KRules Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from rx import subject as rx_subject
from dependency_injector import providers
from krules_core import RuleConst
from krules_core.base_functions.misc import PyCall

from krules_core.core import RuleFactory

from krules_core.providers import (
    event_router_factory,
    proc_events_rx_factory,
    subject_factory,
    subject_storage_factory)

from .. import get_value_from_payload_diffs

counter = 0
asserted = []


@pytest.fixture
def subject():
    global counter
    counter += 1

    return subject_factory('test-subject-{0}'.format(counter)).flush()


@pytest.fixture
def router():
    router = event_router_factory()
    router.unregister_all()
    proc_events_rx_factory.queue.clear()

    return event_router_factory()


@pytest.fixture
def asserted():
    global asserted
    asserted = []
    return asserted


filters = RuleConst.FILTERS
processing = RuleConst.PROCESSING
rulename = RuleConst.RULENAME
passed = RuleConst.PASSED


def _assert(name, expr, msg="test failed"):
    global asserted
    assert expr, msg
    asserted.append(name)


def test_pycall(subject, router, asserted):

    def _func(in_value, raise_error, exc_to_raise=Exception("imbad")):

        if raise_error:
            raise exc_to_raise
        return in_value

    RuleFactory.create(
        "test-pycall-with-error",
        subscribe_to="test-pycall",
        data={
            processing: [
                PyCall(_func, ([1, 2], True),
                       on_success=lambda self: lambda x: x.reverse(),
                       on_error=lambda self: lambda x: self.payload.update({"got_errors": True}))
            ]
        }
    )

    RuleFactory.create(
        "test-pycall-no-error",
        subscribe_to="test-pycall",
        data={
            processing: [
                PyCall(_func, ([1, 2],), kwargs={"raise_error": False},
                       on_success=lambda self: lambda x: (
                           x.reverse(),
                           self.payload.update({"got_errors": False})
                       ))
            ]
        }
    )

    proc_events_rx_factory.subscribe(
        lambda x: x[rulename] == "test-pycall-no-error" and _assert(
            x[rulename],
            get_value_from_payload_diffs("pycall_returns", x[processing][0]["payload_diffs"], default_value=None) == [2, 1]
            and not get_value_from_payload_diffs("got_errors", x[processing][0]["payload_diffs"], default_value=False)
        )
    )

    proc_events_rx_factory.subscribe(
        lambda x: x[rulename] == "test-pycall-with-error" and _assert(
            x[rulename],
            not get_value_from_payload_diffs("pycall_returns", x[processing][0]["payload_diffs"], default_value=None) and
            get_value_from_payload_diffs("got_errors", x[processing][0]["payload_diffs"], default_value=False)
        )
    )

    router.route("test-pycall", subject, {})

    assert "test-pycall-no-error" in asserted
    assert "test-pycall-with-error" in asserted